#!/usr/bin/env python
import sys, re, optparse
from ctypes import *

"""
This script will use the prototypes from "checkdocs.py -s" to concoct
a 1:1 Python wrapper for Allegro.
"""

class Allegro:
    def __init__(self):
        self.types = {}
        self.functions = {}
        self.constants = {}

    def add_struct(self, name):
        x = type(name, (Structure, ), {})
        self.types[name] = x

    def add_union(self, name):
        x = type(name, (Union, ), {})
        self.types[name] = x

    def get_type(self, ptype):
        conversion = {
            "bool" : c_bool,
            "_Bool" : c_bool,
            "char" : c_byte,
            "unsignedchar" : c_ubyte,
            "int" : c_int,
            "unsigned" : c_uint,
            "unsignedint" : c_uint,
            "int16_t" : c_int16,
            "int32_t" : c_int32,
            "uint32_t" : c_uint32,
            "int64_t" : c_int64,
            "uint64_t" : c_uint64,
            "uintptr_t" : c_void_p,
            "intptr_t" : c_void_p,
            "GLuint" : c_uint,
            "unsignedlong" : c_ulong,
            "long" : c_long,
            "size_t" : c_size_t,
            "off_t" : c_int64,
            "time_t" : c_int64,
            "va_list" : c_void_p,
            "float" : c_float,
            "double" : c_double,
            "al_fixed" : c_int,
            "HWND" : c_void_p,
            
            # hack: this probably shouldn't be in public the docs
            "postprocess_callback_t" : c_void_p,
            }
        
        ptype = re.sub(r"\bconst\b", "", ptype)
        ptype = re.sub(r"\extern\b", "", ptype)
        ptype = re.sub(r"\__inline__\b", "", ptype)
        ptype = re.sub(r"\s+", "", ptype)

        if ptype.endswith("*"):
            return c_void_p
        elif ptype in self.types:
            return self.types[ptype]
        else:
            try:
                return conversion[ptype]
            except KeyError:
                print("Error:" + str(ptype))
        return None

    def parse_funcs(self, funcs):
        """
        Go through all documented functions and add their prototypes
        as Python functions.
        
        The file should have been generated by Allegro's documentation
        generation scripts.
        """

        for func in funcs:
            name, proto = func.split(":")
            proto = proto.strip()
            name = name[:-2]
            if proto.startswith("enum"): continue
            if proto.startswith("typedef"): continue
            if "=" in proto: continue
            if proto.startswith("#"): continue
            funcstart = proto.find(name)
            funcend = funcstart + len(name)
            ret = proto[:funcstart].rstrip()
            params = proto[funcend:].strip(" ;")
            if params[0] != "(" or params[-1] != ")":
                print("Error:")
                print(params)
                continue
            params2 = params[1:-1]
            # remove callback argument lists
            balance = 0
            params = ""
            for c in params2:
                if c == ")": balance -= 1
                if balance == 0:
                    params += c
                if c == "(": balance += 1
            params = params.split(",")
            plist = []
            for param in params:
                param = re.sub(r"\bconst\b", "", param)
                param = param.strip()
                if param == "void": continue
                if param == "": continue
                if param == "...": continue

                # treat arrays as a void pointer, for now
                if param.endswith("]") or param.endswith("*"):
                    plist.append(c_void_p)
                    continue
                
                # treat callbacks as a void pointer, for now
                if param.endswith(")"):
                    plist.append(c_void_p)
                    continue

                mob = re.match("^.*?(\w+)$", param)
                if mob:
                    pnamepos = mob.start(1)
                else:
                    print(params)
                    print(proto)
                    print("")
                    continue
                ptype = param[:pnamepos]
                ptype = self.get_type(ptype)
                plist.append(ptype)

            f = type("", (object, ), {"restype": c_int})
            if not ret.endswith("void"):
                f.restype = self.get_type(ret)
            try:
                f.argtypes = plist
            except TypeError, e:
                print(e)
                print(name)
                print(plist)
            self.functions[name] = f

    def parse_protos(self, filename):
        protos = []
        unions = []
        funcs = []

        # first pass: create all structs, but without fields
        for line in open(filename):
            name, proto = line.split(":")
            proto = proto.lstrip()
            if name.endswith("()"):
                funcs.append(line)
                continue
            # anonymous structs have no name at all
            if name and not name.startswith("ALLEGRO_"): continue
            if name == "ALLEGRO_OGL_EXT_API": continue
            if proto.startswith("union") or\
                proto.startswith("typedef union"):
                self.add_union(name)
                unions.append((name, proto))
            elif proto.startswith("struct") or\
                proto.startswith("typedef struct"):
                self.add_struct(name)
                protos.append((name, proto))
            elif proto.startswith("enum") or\
                proto.startswith("typedef enum"):
                if name: self.types[name] = c_int
                protos.append(("", proto))
            else:
                # actual typedef
                mob = re.match("typedef (.*) " + name, proto)
                if mob:
                    t = mob.group(1)
                    self.types[name] = self.get_type(t.strip())
                else:
                    # Probably a function pointer
                    self.types[name] = c_void_p

        # Unions must come last because they finalize the fields.
        protos += unions

        # second pass: fill in fields
        for name, proto in protos:
            bo = proto.find("{")
            if bo == -1:
                continue
            bc = proto.rfind("}")
            braces = proto[bo + 1:bc]
           
            if proto.startswith("enum") or \
                proto.startswith("typedef enum"):
                
                fields = braces.split(",")
                i = 0
                for field in fields:
                    if "=" in field:
                        fname, val = field.split("=", 1)
                        fname = fname.strip()
                        i = int(eval(val, globals(), self.constants))
                    else:
                        fname = field.strip()
                    if not fname: continue
                    self.constants[fname] = i
                    i += 1
                continue

            balance = 0
            fields = [""]
            for c in braces:
                if c == "{": balance += 1
                if c == "}": balance -= 1
                if c == ";" and balance == 0:
                    fields.append("")
                else:
                    fields[-1] += c

            flist = []
            for field in fields:
                if not field: continue
                
                # add function pointer as void pointer
                mob = re.match(".*?\(\*(\w+)\)", field)
                if mob:
                    flist.append((mob.group(1), c_void_p))
                    continue
                
                # add any pointer as void pointer
                mob = re.match(".*?\*(\w+)$", field)
                if mob:
                    flist.append((mob.group(1), c_void_p))
                    continue
                
                # add any array as void pointer
                mob = re.match(".*?(\w+)\[.*\]$", field)
                if mob:
                    flist.append((mob.group(1), c_void_p))
                    continue
                
                vars = field.split(",")
                mob = re.match("\s*(.*?)\s+(\w+)\s*$", vars[0])
                t = self.get_type(mob.group(1))
                flist.append((mob.group(2), t))
                for v in vars[1:]:
                    flist.append((v.strip(), t))

            try: self.types[name]._fields_ = flist
            except AttributeError:
                print(name, flist)
                
        self.parse_funcs(funcs)

def main():
    p = optparse.OptionParser()
    p.add_option("-o", "--output", help = "location of generated file")
    p.add_option("-p", "--protos", help = "A file with all " +
        "prototypes to generate Python wrappers for, one per line. "
        "Generate it with docs/scripts/checkdocs.py -p")
    options, args = p.parse_args()
    
    if not options.protos:
        p.print_help()
        return

    al = Allegro()

    al.parse_protos(options.protos)

    f = open(options.output, "w") if options.output else sys.stdout

    f.write(r"""# Generated by generate_python_ctypes.py.
import os
from ctypes import *
from ctypes.util import *

# You must adjust this function to point ctypes to the A5 DLLs you are
# distributing.
_dlls = []
def _add_dll(name):
    release = "-debug"
    version = "-4.9.14"

    path = find_library(name + release + version)
    if not path:
		if os.name == "mac":
			path = name + release + version + ".dylib"
		elif os.name == "nt":
			path = name + release + version + ".dll"
		elif os.name == "posix":
			path = "lib" + name + release + version + ".so"
		else:
			sys.stderr.write("Cannot find library " + name + "\n")
 
		# In most cases, you actually don't want the above and instead
		# use the exact filename within your game distribution, possibly
		# even within a .zip file. Here we simply look inside a folder
		# relative to the current path...
		if not path:
			path = "dlls/" + path

    _dlls.append(CDLL(path))

_add_dll("allegro")
_add_dll("allegro_audio")
_add_dll("allegro_flac")
_add_dll("allegro_primitives")
_add_dll("allegro_color")
_add_dll("allegro_font")
_add_dll("allegro_ttf")
_add_dll("allegro_image")
_add_dll("allegro_vorbis")
_add_dll("allegro_dialog")
_add_dll("allegro_memfile")
_add_dll("allegro_physfs")

# We don't have information ready which A5 function is in which DLL,
# so we just try them all.
def _dll(func, ret, params):
    for dll in _dlls:
        try:
            f = dll[func]
            f.restype = ret
            f.argtypes = params
            return f
        except AttributeError: pass
    sys.stderr.write("Cannot find function " + func + "\n")
    return lambda *args: None
""")

    for name, x in sorted(al.types.items()):
        base = x.__bases__[0]
        if base != Structure and base != Union:
            f.write(name + " = " + x.__name__ + "\n")

    for kind in Structure, Union:
        for name, x in sorted(al.types.items()):
            base = x.__bases__[0]
            if base != kind: continue
            f.write("class " + name + "(" + base.__name__ + "): pass\n")

        for name, x in sorted(al.types.items()):
            base = x.__bases__[0]
            if base != kind: continue
            if hasattr(x, "_fields_"):
                f.write(name + "._fields_ = [\n")
                for name, type in x._fields_:
                    f.write("    (\"" + name + "\", " + type.__name__ + "),\n")
                f.write("    ]\n")

    for name, x in sorted(al.functions.items()):
        f.write(name + " = _dll(\"" + name + "\", ")
        f.write(x.restype.__name__ + ", ")
        f.write("[" + (", ".join([a.__name__ for a in x.argtypes])) +
            "])\n")

    for name, val in sorted(al.constants.items()):
        f.write(name + " = " + str(val) + "\n")

    f.close()

main()
